%for part 2 a kalman filter will be developed for tracking a target moving
%in 2D. The targets motion is defined by kinematic equations with a small
%amount of random motion captured by the acceleration noise as will become
%apparent below, and the sensors are corrupted by additive white gaussian
%noise. the result should demonstrate the filters effectiveness
clear all;
clc;
clf;
N = 500;
Ts = 0.02; %50Hz sampling
Time = 1:1:N;
Time = Time*Ts;

A = [0 1 0 0;
    0 0 0 0;
    0 0 0 1;
    0 0 0 0];

B = [0 1 0 0;
    0 0 0 0.3]'; %use this to input an initial velocity

C = [1 0 0 0;
    0 0 1 0];

D = 0;
sys = ss(A,B,C,D);
sysd = c2d(sys,Ts); %Discretise the continuous ss sys


input = [zeros(5,2);ones(N-5,2)]';


Q = [0 0 0 0;
    0 0.1 0 0;
    0 0 0 0;
    0 0 0 0.1]; %Covariance state
Q = Q + ones(4)*10^-10;
WQ = Q^(-1/2);
DWQ = inv(WQ);

noiseState = DWQ*randn(4,N);

R = [7 0;
    0 7]; %covariance sensors
R = R + ones(2)*10^-10;
WR = R^(-1/2); %take a variable with covar mat R to a whitened state
DWR = inv(WR); %dewhiten a white unit power GRV to covar R


noiseSensors = DWR*randn(2,N); %producing noise with the given covariance


Kalman = dsp.KalmanFilter(sysd.A,sysd.C,Q,R,sysd.B,...
            'InitialStateEstimate', [0;0;0;0],...
            'InitialErrorCovarianceEstimate',[0 0 0 0;0 0 0 0;0 0 0 0;0 0 0 0],...
            'ControlInputPort',false); %%this is the kalman filter object. implementing it from scratch is shown below

state = [0;0;0;0]; %True inital state
for i = 2:1:N %%motion simulation
    
    state(:,i) = sysd.A*state(:,i-1) + sysd.B*input(:,i-1) + noiseState(:,i-1);
    Measurement(:,i) = sysd.C*state(:,i-1) + noiseSensors(:,i-1);
    estVal(:,i) = step(Kalman, Measurement(:,i));
    
end

figure(1)
hold on
plot(Time,Measurement');
plot(Time,state([1,3],:)','k');

plot(Time,estVal','b--');










%%Direct implementation optimal gain KF - directly from wikipedia :D
%this implementation is important to the Extended Kalman filter which
%linearises a non-linear system about the local state at each step

estimatedState = [2;-3;1;2]; %%inital state estimate
currentCovariance = [1 0 0 0;0 1 0 0;0 0 1 0;0 0 0 1];

for i = 2:1:N 
    predictedState = sysd.A*estimatedState(:,i-1); %predict positon based on previous state estimate and control effort if known/applicable
    PredictedP = sysd.A*currentCovariance*sysd.A' + Q;
    
    innovation = Measurement(:,i) - sysd.C*predictedState;
    InnovationCov = sysd.C*PredictedP*(sysd.C)' + R;
    KalGain = PredictedP*sysd.C'*inv(InnovationCov);
    Kgain(i,:) = [KalGain(:,1);KalGain(:,2)];
    estimatedState(:,i) = predictedState + KalGain*innovation;
    currentCovariance = (eye(4) - KalGain*sysd.C)*PredictedP;   
end

figure(2)
hold on;
plot(Time,Measurement');
plot(Time,state([1,3],:)','k');
plot(Time,(sysd.C*estimatedState)','b--');


figure(3)
plot(Time,Kgain) %note that the kalman gain converges to a steady state. This is typical. Computing the Kalman Gain is expensive as it has a matrix inverse. Generally its better to find the steady state gain and fix the terms. see below
%For the eqn that computes the steady state gain


%%Steady state KF
Lupanyov = zeros(4);
for i =1:1:1000
    Lupanyov = sysd.A*Lupanyov*sysd.A' + Q - sysd.A*Lupanyov*sysd.C'*inv(sysd.C*Lupanyov*sysd.C' + R)*sysd.C*Lupanyov*sysd.A';
end
Gain = sysd.A*Lupanyov*sysd.C'*inv(sysd.C*Lupanyov*sysd.C' + R);


%now the predict->update is much faster:
for i = 2:1:N 
    predictedState = sysd.A*estimatedState(:,i-1); 
    
    innovation = Measurement(:,i) - sysd.C*predictedState;
    estimatedState(:,i) = predictedState + Gain*innovation;  
end

figure(4)
hold on
plot(Time,Measurement');
plot(Time,state([1,3],:)','k');
plot(Time,(sysd.C*estimatedState)','b--');